#include "ColorKarma.h"
#include "GrafUtils.h"
#include "CoMation.h"
#include "MenuDispatch.h"
#include "GameSounds.h"
#include "xthing.h"
#include "GlobalDefs.h"

#define kCopyBitsMethod 1
#define kScrollRectMeth 2

#define kMusic 		128

#define kDude  		129
#define kWhistle 	130
#define kCool		131
#define kVikingMind	132
#define kMaxSound	4

static WindowPtr	gKarmaWindow;			// window to scroll it into
static GWorldPtr	gKarmaWorld;			// off-screen image to be scrolled
static RgnHandle	gKarmaMask;				// masking area to be transferred

static GWorldPtr	gOrbWorld;
static RgnHandle	gOrbClip;

static Rect			gKarmaBounds;			// area of offscreen that appears in window
static Point		gOrbVel;				// 
static Boolean		gConeVis;				// is the orb vis?
static Boolean		gColorCycle;			// are we animating the palette now?
static Boolean		gSpeedEval;
static OSErr		err;					// so don't have to declare everywhere
static short		gKarmaWindowSize;		// indicates which window resource to use		
static Rect			gOrbBounds;				// position of the orb on the screen
static long			gKarmaScrollSpeed;		// scrolling speed;
static short		gCurSpeedItem;			// currently selected speed (pixels moved)
static short		gAnimationMethod;		// copybits or scrollrect?

static xthing		gCycleTask;				// used for timing palette animation
static CTabHandle	gAnimTable;				// used for palette animation

static xthing		gSpeedTask;				// used for timing FPS calculations
static long			gFrameCount;			// for calculating FPS
static xthing		gSoundTask;

static RgnHandle	gUtilRgn;
static RgnHandle	gSaveRgn;

void 	CenterOrb(void);
void 	DrawKarmaWindow(void);
void	ScrollKarmaWindow(Point);


void ToggleCone(void)
/*
	toggles the cone on and off
*/
{
	gConeVis = !gConeVis;
	if(gConeVis)
	{
		SetItemMark( GetMHandle(KARMA_MENU), KARMA_CONE, diamondMark );
		CenterOrb();
	}
	else
	{
		SetItemMark( GetMHandle(KARMA_MENU), KARMA_CONE, noMark );
		
		if(gKarmaMask)	DisposeRgn(gKarmaMask);
		gKarmaMask = NewRgn();
		OpenRgn();
		FrameRect(&gKarmaWindow->portRect);
		CloseRgn(gKarmaMask);
		DrawKarmaWindow();
	}
	
}


Boolean PlaySoundTask(xthing *xtp, long dataRef)
{
	OSErr	err;
	short   r,c;
	
	
	if( (Random() & 0x00fF) > 128 )
	{
		/* play a random sound on a random channel */
		r = (Random() & 0x00ff ) % kMaxSound;
		c = (Random() & 0x0ff) % 3;
		
		err = PlaySndAsynchChannel( 129 + r, 1 + c, kStdPriority);
	}
		
	return true;
}

Boolean SpeedEvaluationTask(xthing *xtp, long dataRef)
{
	DialogPtr	dp;
	Str32		speed, seconds, numFrames;
	short		item;
	
	/* when this task completes, it means timing is complete, */
	/* all we have to do is display a dialog and return false */
	/* so xthing won't start the task over again */
		
	NumToString(gFrameCount / 10,speed);
	NumToString(10, seconds);
	NumToString(gFrameCount, numFrames);
	
	dp = GetNewDialog(200, nil, (DialogPtr)-1);
	ParamText(seconds,speed,numFrames,nil);
	ModalDialog(nil,&item);
	DisposeDialog(dp);
	gSpeedEval = false;

	return false;
}

void StartAnimation(void)
{
	gOrbVel.h = 8;
	gOrbVel.v = 0;
}

void StopAnimation(void)
{
	gOrbVel.h = 0;
	gOrbVel.v = 0;
}

Boolean AnimateTask(void)
{
	Point	saveVel;
	
	gFrameCount++;
	
	saveVel = gOrbVel;
	if(gOrbVel.v || gOrbVel.h)
	{
		OffsetRect(&gKarmaBounds, gOrbVel.h, gOrbVel.v);
	
		if(gKarmaBounds.top < gKarmaWorld->portRect.top)
		{
			saveVel.v = saveVel.v - (gKarmaWorld->portRect.top - gKarmaBounds.top);
			gKarmaBounds.bottom = gKarmaWorld->portRect.top + gKarmaBounds.bottom - gKarmaBounds.top;
			gKarmaBounds.top = gKarmaWorld->portRect.top;
			gOrbVel.v = -gOrbVel.v;
		} 
		else if(gKarmaBounds.bottom > gKarmaWorld->portRect.bottom)
		{
			saveVel.v = saveVel.v - (gKarmaBounds.bottom - gKarmaWorld->portRect.bottom);
			gKarmaBounds.top = gKarmaWorld->portRect.bottom - (gKarmaBounds.bottom - gKarmaBounds.top);
			gKarmaBounds.bottom = gKarmaWorld->portRect.bottom;
			gOrbVel.v = -gOrbVel.v;
		}

		if(gKarmaBounds.left < gKarmaWorld->portRect.left)
		{
			saveVel.h = saveVel.h - (gKarmaWorld->portRect.left - gKarmaBounds.left);
			gKarmaBounds.right = gKarmaWorld->portRect.left + gKarmaBounds.right - gKarmaBounds.left;
			gKarmaBounds.left = gKarmaWorld->portRect.left;
			gOrbVel.h = -gOrbVel.h;
		}
		else if(gKarmaBounds.right > gKarmaWorld->portRect.right)
		{
			saveVel.h = saveVel.h - (gKarmaBounds.right - gKarmaWorld->portRect.right);
			gKarmaBounds.left = gKarmaWorld->portRect.right - (gKarmaBounds.right - gKarmaBounds.left);
			gKarmaBounds.right = gKarmaWorld->portRect.right;
			gOrbVel.h = -gOrbVel.h;
		}
		if(gAnimationMethod == kCopyBitsMethod)
			DrawKarmaWindow();
		else
			ScrollKarmaWindow(saveVel);
	}
	
	
	return true;
}

Boolean CycleTask(xthing *xtp, long dataRef)
{
	short	i,swap;
	PaletteHandle pal;
	// animate all the palette entries by bumping up
	
	if(gAnimTable)
	{
		for(i = 0; i <= (**gAnimTable).ctSize; i++)
		{
			swap = i+1;
			if(i >(**gAnimTable).ctSize) i = 0;
			(**gAnimTable).ctTable[i] = (**gAnimTable).ctTable[swap];
			/*
			(**gAnimTable).ctTable[i].rgb.red   +=1024; 
			(**gAnimTable).ctTable[i].rgb.green +=1024; 
			(**gAnimTable).ctTable[i].rgb.blue  +=1024; 
			*/
		}
	
		// IM-5 164		
		AnimatePalette(gKarmaWindow, gAnimTable, 0, 0, (**gAnimTable).ctSize + 1);
	}
	
	if(!gColorCycle)		/* restore palette */
	{
		pal = GetPalette(gKarmaWindow);
		// restore the palette back to tolerant when done animating
		for(i = 0; i < (**pal).pmEntries; i++)
		{
			SetEntryUsage(pal,i, pmTolerant, 0);
		}
		ActivatePalette(gKarmaWindow);
		DisposCTable( gAnimTable );
		gAnimTable = nil;
		/*
		pal = GetNewPalette(5000);
		SetPalette(gKarmaWindow, pal, true);
		ActivatePalette(gKarmaWindow);
		*/
	}

	/* keep going until the user de-selects the menu item */
	return gColorCycle;
}

void ToggleCycle(void)
{
	PaletteHandle	pal;
	short	i;
	
	gColorCycle = !gColorCycle;
	if(gColorCycle)
	{
		SetItemMark( GetMHandle(KARMA_MENU), KARMA_CYC, diamondMark );
		
		/* set up the palette for animation */
		pal = GetPalette(gKarmaWindow);					// IM-5 pg 163			
		for(i = 0; i < (**pal).pmEntries; i++)
		{
			// make the entry animated.  Passing -1 for tolerance 
			// means don't change the tolerance.
			SetEntryUsage(pal,i, pmAnimated , -1);		// IM-5 pg 165
		}
		ActivatePalette(gKarmaWindow);					// IM-5 pg 162
		
		
		gAnimTable = GetCTable(8);
		Palette2CTab(pal, gAnimTable);
		/* start the palette animation task for, 2 times a second */
		StartXThing(&gCycleTask, 500, CycleTask, 0);
	}
	else
	{
		SetItemMark( GetMHandle(KARMA_MENU), KARMA_CYC, noMark );
		
	}
	
}

void SpeedTest()
{
	if(gSpeedEval) return;
	
	gSpeedEval = true;
	gFrameCount = 0;
	StartXThing(&gSpeedTask, 1000 * 10, SpeedEvaluationTask, 0);
	
	
}

void PlayMusic()
{
	err =  PlaySndAsynchChannel(128, kChan1, kStdPriority);
	if(err != noErr)
		ErrMsgCode("\p error playing sound",err);

	StartXThing(&gSoundTask, 2000, PlaySoundTask, 0);

}

void ChooseAnimMeth(short item)
{
	CheckItem( GetMHandle(ANIM_METH), gAnimationMethod,  false);
	gAnimationMethod = item;
	CheckItem( GetMHandle(ANIM_METH), gAnimationMethod, true);
}

void ChooseAnimSpeed(short item)
{
	Str255	itemString;
	
	CheckItem( GetMHandle(ANIM_SPEED), gCurSpeedItem,  false);
	CheckItem(GetMHandle(ANIM_SPEED), item, true);

	gCurSpeedItem = item;
	GetItem( GetMHandle(ANIM_SPEED), item, itemString );
	StringToNum(itemString, &gKarmaScrollSpeed);
}

void InitKarmaGlobals()
{
	gKarmaWindowSize = 0;
	gKarmaWindow = nil;
	gKarmaWorld = nil;
	gOrbClip = nil;
	gKarmaMask = nil;
	err = noErr;
	gKarmaScrollSpeed = 8;
	gCurSpeedItem = 4;
	gAnimationMethod = 1;
	gConeVis = true;
	gColorCycle = false;
	gAnimTable = nil;
	gUtilRgn = NewRgn();
	gSaveRgn = NewRgn();
	gSpeedEval = false;
}

void DrawOrb(void)
{
	PixMapHandle	src,dst;
	
	if(gConeVis)
	{
		src = PreserveGraf(gOrbWorld);
		dst = PreserveGraf(gKarmaWindow);
		
		CopyBits(*src, *dst, &gOrbWorld->portRect, &gOrbBounds, srcCopy, gOrbClip);
		
		RestoreGraf();
		RestoreGraf();
	}
}

pascal	void  CatchStdRgn( GrafVerb verb, RgnHandle theRgn )
{
	PixMapHandle	src;
	
	if( verb == erase)
	{
		src = GetGWorldPixMap( gKarmaWorld );
		
		LockPixels( src );		// not purgeable, so ignore result		
		CopyBits( *src, &gKarmaWindow->portBits, &gKarmaBounds, &thePort->portRect, srcCopy, theRgn);
		UnlockPixels( src);
		
	} else
		StdRgn(verb, theRgn);
}

void	ScrollKarmaWindow(Point vel)
/*
	scroll the window using scroll rect
	under system 7, if you pass nil to scroll rect for the
	region parameter, it will not erase the
	background as under system 6.
	
	instead of this, I am replacing the erase region bottleneck
	and using CopyBits from there, instead of letting scroll rect erase it.
	
	My testing shows this to be slower than CopyBits alone.
*/
{
	CQDProcs		cProcs;
	PixMapHandle	src; 		
	PixMapHandle	dst;
	
	dst = PreserveGraf(gKarmaWindow);
	SetStdCProcs(&cProcs);
	cProcs.rgnProc = CatchStdRgn;
	gKarmaWindow->grafProcs = &cProcs;
	
	if(gConeVis)
	{
		GetClip(gSaveRgn);
		SetClip(gKarmaMask);
	}
	
	if(gConeVis)
		SetClip(gSaveRgn);
		
	RestoreGraf();
	

	gKarmaWindow->grafProcs = nil;
}

void 	DrawKarmaWindow(void)
{
	PixMapHandle	src, dst;


	if(!gPrinting)
	{
		
		if(!gKarmaWindow) return;
		
		/* update background */
		src = PreserveGraf(gKarmaWorld);
		dst = PreserveGraf(gKarmaWindow);
		
		CopyBits (*src, *dst, &gKarmaBounds, &gKarmaWindow->portRect, srcCopy, gKarmaMask);
		
		
		RestoreGraf();
		RestoreGraf();
	}
	else							/* we are printing so, do it right... */
	{
		GrafPtr		savePort;
		
		
		GetPort(savePort);								// get the print port
		src = PreserveGraf(gKarmaWorld);
		dst = PreserveGraf(savePort);
		
		CopyBits (*src, *dst, &gKarmaBounds, &savePort->portRect, srcCopy, nil);
		
		RestoreGraf();
		RestoreGraf();

	}
}

void CenterOrb(void)
{
	BitMap	maskMap;
	
	/* set up the position of the orb, in the center of the window */
	gOrbBounds.top = (gKarmaWindow->portRect.bottom / 2) - (RECT_HT(gOrbWorld->portRect) / 2);
	gOrbBounds.left = (gKarmaWindow->portRect.right / 2) - (RECT_WD(gOrbWorld->portRect) / 2);
	gOrbBounds.bottom = gOrbBounds.top + RECT_HT(gOrbWorld->portRect);
	gOrbBounds.right = gOrbBounds.left + RECT_WD(gOrbWorld->portRect);
		
	if(gOrbClip) DisposeRgn(gOrbClip);
		
	/* read in the mask pict, load to a BitMap */
	LoadPicture(130,&maskMap);
	if(!maskMap.baseAddr)
		ErrMsg("\pError allocating bitmap for hex mask");
	
	/* Create a region from that bitmap */
	gOrbClip = NewRgn();
	err =  BitMapToRegion(gOrbClip,&maskMap);
	if(err) ErrMsgCode("\p BitMapToRegion [1]",err);
	FreeBitMap(&maskMap);


	OffsetRgn(gOrbClip, gOrbBounds.left, gOrbBounds.top);
	if(gKarmaMask)	DisposeRgn(gKarmaMask);
	
	gKarmaMask = NewRgn();
	OpenRgn();
	FrameRect(&gKarmaWindow->portRect);
	CloseRgn(gKarmaMask);
	
	DiffRgn(gKarmaMask, gOrbClip, gKarmaMask);
	DrawOrb();
}

void CreateKarmaWindow(void)
{
	PaletteHandle 	pal;
	CTabHandle	 	gwctab;
	short			i;
	Rect			tweenBounds;
	
	gKarmaWindow = GetNewCWindow(gKarmaWindowSize + 128, nil, (WindowPtr)-1);

	gwctab = GetCTable(8);
	SetPort(gKarmaWindow);
	pal = GetNewPalette(5000);
	SetPalette(gKarmaWindow, pal, true);
	ActivatePalette(gKarmaWindow);

	
	err = CreateGWorldFromPictResource(&gKarmaWorld, 128);
	if(err) {
		ErrMsgCode("\pNewGWorld failed!",err);
		ExitToShell();
	}
	
	gwctab = (**GetGWorldPixMap(gKarmaWorld)).pmTable;	// get the color table for the gWorld.
	
	// Set up the clut so that the entries refer to palette
	// indexes, instead of the usual rgb information.
	(**gwctab).ctFlags |= 0x4000;		// this bit is the flag to indicate this
	// See Inside Macintosh Volume VI page 20-17


	for (i = 0; i <= (**gwctab).ctSize; i++)
	{
		(**gwctab).ctTable[i].value = i;
	}
	CTabChanged(gwctab);

	/* change the color table seed to match the screen seed */
	/* this prevents any color matching, and is generally fun code to write */
	(**gwctab).ctSeed = (**(**(*(CGrafPtr)gKarmaWindow).portPixMap).pmTable).ctSeed;

	/* Load in the orb object */	 
	err = CreateGWorldFromPictResource(&gOrbWorld, 129);

	/* now do the same things for this ctable */
	gwctab = (**GetGWorldPixMap(gOrbWorld)).pmTable;	// get the color table for the gWorld.
	(**gwctab).ctFlags |= 0x4000;		// this bit is the flag to indicate this
	for (i = 0; i <= (**gwctab).ctSize; i++)
	{
		(**gwctab).ctTable[i].value = i;
	}
	CTabChanged(gwctab);
	(**gwctab).ctSeed = (**(**(*(CGrafPtr)gKarmaWindow).portPixMap).pmTable).ctSeed;

	/* initial area in view */
	gKarmaBounds = gKarmaWindow->portRect;
	CenterOrb();
	DrawKarmaWindow();

	ValidRect(&gKarmaWindow->portRect);

	/* start the animation task */
	//StartXThing(&gAnimTask, 10, AnimateTask, 0);
}

void		ResizeKarmaWindow(void)
{
	short	saveT, saveL;
	PaletteHandle 	pal;
	
	if(!gKarmaWindow) return;
	
	gKarmaWindowSize = (gKarmaWindowSize + 1) % kMaxWindowSize;
	DisposeWindow(gKarmaWindow);
	gKarmaWindow = GetNewCWindow(gKarmaWindowSize + 128, nil, (WindowPtr)-1);
	pal = GetNewPalette(5000);
	SetPalette(gKarmaWindow, pal, true);
	ActivatePalette(gKarmaWindow);

	saveT = gKarmaBounds.top;
	saveL = gKarmaBounds.left;
	
	gKarmaBounds = gKarmaWindow->portRect;
	
	OffsetRect(&gKarmaBounds, saveT, saveL);
	
	if(gColorCycle)
	{
		RmvTime( &gCycleTask.timer );
		gColorCycle = false;
		ToggleCycle();
	}
	
	if(gConeVis)
		CenterOrb();
	DrawKarmaWindow();
}

void		DisposeKarmaWindow(void)
{
	if(!gKarmaWindow) return;
	
	DisposeWindow(gKarmaWindow);
	DisposeGWorld(gKarmaWorld);
}


void		DoKarmaKey(EventRecord *event)
{
	char	k;
	Rect	oldPos;
	Boolean	needDraw = false;
		
	k = event->message & charCodeMask;
	
	if(gAnimationMethod == kCopyBitsMethod)
	{
		
		switch(k)
		{
			case	'i':								/* up */
			case	'I':
			case	'8':	gOrbVel.v += -gKarmaScrollSpeed;
							break;
			case	'k':								/* down */
			case	'K':
			case	'2':	gOrbVel.v += gKarmaScrollSpeed;
							break;
			case	'j':								/* left */
			case	'J':
			case	'4':	gOrbVel.h += -gKarmaScrollSpeed;
							break;
			case	'l':								/* right */
			case	'L':
			case	'6':	gOrbVel.h += gKarmaScrollSpeed;
							break;
		}


	} 
	else
	{
	
		switch(k)
		{
			case	'i':								/* up */
			case	'I':
			case	'8':	gOrbVel.v += -gKarmaScrollSpeed;
							gOrbVel.h = 0;
							break;
			case	'k':								/* down */
			case	'K':
			case	'2':	gOrbVel.v += gKarmaScrollSpeed;
							gOrbVel.h = 0;
							break;
			case	'j':								/* left */
			case	'J':
			case	'4':	gOrbVel.h += -gKarmaScrollSpeed;
							gOrbVel.v = 0;
							break;
			case	'l':								/* right */
			case	'L':
			case	'6':	gOrbVel.h += gKarmaScrollSpeed;
							gOrbVel.v = 0;
							break;
		}
	}
	
	if(gOrbVel.h > 64)	gOrbVel.h = 64;
	if(gOrbVel.h < -64) gOrbVel.h = -64;
	
	if(gOrbVel.v > 64) gOrbVel.v = 64;
	if(gOrbVel.v < -64) gOrbVel.v = 64;
	


}	

void		DoKarmaClick(EventRecord *event)
{
	
}

void DoKarmaIdleTask(void)
{
	SoundKeeper();
	ProcessXThingTask(0);
	AnimateTask();
}


